from dataclasses import dataclass, asdict
from appyter.ext.urllib import join_url, parent_url

@dataclass(frozen=True,)
class FileInfo:
  file: str
  filename: str
  metadata: dict = None

async def add_file(data: FileInfo, auth=None, config=None):
  import aiohttp
  async with aiohttp.ClientSession(
    headers={
      'Authorization': f"Bearer {auth}",
    } if auth else {},
    raise_for_status=True,
  ) as session:
    async with session.post(
      join_url(
        parent_url(config['PUBLIC_URL']),
        'postgrest/rpc/add_file',
      ),
      json=asdict(data),
    ) as res:
      return await res.json()

async def list_files(auth=None, config=None):
  if not auth: raise PermissionError
  import aiohttp
  async with aiohttp.ClientSession(
    headers={
      'Authorization': f"Bearer {auth}",
    },
    raise_for_status=True,
  ) as session:
    async with session.get(
      join_url(
        parent_url(config['PUBLIC_URL']),
        'postgrest/user_file',
      ),
    ) as res:
      return await res.json()



from appyter.ext.contextlib import ensure_context
from appyter.ext.asyncio.helpers import ensure_async

@ensure_async
def fsspec_read_and_run(path, op):
  from appyter.ext.fsspec.core import url_to_fs_ex
  fs, fo = url_to_fs_ex(path)
  with ensure_context(fs) as fs:
    with fs.open(fo, 'r') as fr:
      return op(fr)

@ensure_async
def fsspec_cp(left_fs, left_fo, right_fs, right_fo):
  import shutil
  from appyter.ext.urllib import parent_url
  with left_fs.open(left_fo, 'rb') as fr:
    right_fs.makedirs(parent_url(right_fo), exist_ok=True)
    with right_fs.open(right_fo, 'wb') as fw:
      shutil.copyfileobj(fr, fw)



''' This module enables you to turn your jupyter notebook into a jinja2 template-driven web application. Or just parse for other purposes.
'''
import os
from appyter.magic import init
from importlib.metadata import version

__version__ = version(__package__)



import json

def try_json_loads(v):
  try:
    return json.loads(v)
  except TypeError:
    return v
  except json.JSONDecodeError:
    return v

def try_json_dumps(v):
  if type(v) == str:
    return v
  else:
    return json.dumps(v)

from appyter.ext.asyncio.helpers import ensure_async
async_json_loads = ensure_async(json.loads)
async_json_dumps = ensure_async(json.dumps)



import re
from appyter.ext.itertools import ensure_list
from appyter.ext.re import re_full
from appyter.fields import Field, FieldConstraintException
from appyter.profiles.default.fields.FileField import FileField
from appyter.render.flask_app.upload import upload_from_request

class MultiFileField(FileField):
  def __init__(self, multiple=True, **kwargs):
    super().__init__(multiple=multiple, **kwargs)

  @property
  def raw_value(self):
    if type(self.args['value']) == str and self.args['value']:
      raw_value = self.args['value'].splitlines()
    elif type(self.args['value']) == list or type(self.args['value']) == tuple:
      raw_value = list(self.args['value'])
    else:
      raw_value = []
    return [v for v in raw_value if re.sub(r'\s+', '', v)]

  @property
  def uri(self):
    raise NotImplementedError

  @property
  def uris(self):
    return [
      self.prepare_uri(v)
      for v in self.raw_value
    ]

  def prepare(self, req):
    ''' Convert file in request into URI
    '''
    data_paths = None
    # get data from request upload
    if getattr(req, 'files', None):
      data_paths = upload_from_request(req, self.args['name'], multiple=True)
    # fallback to default (URI in request)
    if data_paths is None:
      data_paths = [s for f in ensure_list(Field.prepare(self, req)[self.args['name']]) for s in f.splitlines()]
    # convert strings into uris
    data_paths = [self.prepare_uri(data_path) for data_path in data_paths]
    data = {self.args['name']: [str(data_path) for data_path in data_paths]}
    for data_path in data_paths:
      if data_path.fragment_path:
        data.update({ f"_file:{data_path.fragment_path}": str(data_path.with_fragment_path(None))})
    return data

  def constraint(self):
    if not self.raw_value:
      return not self.args.get('required')
    else:
      expr = re.compile(re_full(self.args['constraint']))
      for uri in self.uris:
        if self._env.globals['_config']['SAFE_MODE']:
          if uri.scheme not in {'drs', 's3', 'gs', 'ftp', 'http', 'https', 'storage', 'user'}:
            return False
        if not expr.match(uri.fragment_path):
          return False
      return True

  @property
  def value(self):
    if not self.uris:
      return []
    if not self.constraint():
      raise FieldConstraintException(
        field=self.field,
        field_name=self.args['name'],
        value=self.raw_value,
      )
    return [uri.fragment_path for uri in self.uris]

  def to_cwl(self):
    schema = super().to_cwl()
    schema['inputBinding'].pop('prefix', None)
    if self.args.get('required') == True:
      schema['type'] = {
        'type': 'array',
        'items': 'File',
        'inputBinding': {
          'separate': False,
          'prefix': f"--{self.args['name']}="
        },
      }
    else:
      schema['type'] = [
        'null',
        {
          'type': 'array',
          'items': 'File',
          'inputBinding': {
            'separate': False,
            'prefix': f"--{self.args['name']}="
          },
        },
      ]
    return schema

  def to_cwl_value(self):
    return [
      {
        'class': 'File',
        'path': str(uri.with_fragment_path(None)),
        'name': uri.fragment_path,
      } for uri in self.uris
    ] if self.raw_value is not None else None

  def to_click(self):
    args, kwargs = super().to_click()
    kwargs['multiple'] = True
    return args, kwargs

  def to_jsonschema(self):
    schema = super().to_jsonschema()
    schema = {
      'type': 'array',
      'items': schema,
    }
    return schema

  @property
  def public_url(self):
    raise NotImplementedError

  @property
  def public_urls(self):
    return [
      str(uri.with_fragment_path(None))
      for uri in self.uris
    ]



import re
from appyter.ext.urllib import url_filename
from appyter.render.flask_app.constants import get_input_fs
from appyter.ext.fsspec.core import url_to_fs_ex
from appyter.ext.contextlib import ensure_context
from appyter.render.flask_app.upload import organize_file_content
from appyter.context import get_env

def ensure_drs(url):
  ''' Ensures a url is accessible via DRS
  '''
  if url.startswith('storage://input/'):
    return url_filename(url)
  #
  fs, fo = url_to_fs_ex(url)
  try:
    # Some providers, namely SBFS, provide DRS endpoints already
    return fs.get_drs(fo)
  except AttributeError:
    pass
  #
  with ensure_context(fs) as fs:
    storage_id = organize_file_content(get_input_fs(), fs, fo)
  #
  env = get_env()
  drs_id = re.sub(r"^storage://input/", re.sub(r'^https?://', 'drs://', env['PUBLIC_URL'] + '/'), storage_id)
  return drs_id



def re_full(expr):
  if not expr.startswith('^'): expr = '^' + expr
  if not expr.endswith('$'): expr = expr + '$'
  return expr

def slugify(s):
  import re
  return re.sub(r'[^a-zA-Z0-9]+', '_', s)



import uuid
import logging
logger = logging.getLogger(__name__)

from appyter.ext.json import async_json_dumps

class ChunkedEmitMixin:
  ''' A mixin for splitting up large `emit` calls into `chunked` calls of lower priority
  '''
  CHUNK_SIZE = 1<<20 # 1kb

  async def emit(self, evt, data, priority=0, **kwargs):
    chunkable = evt != 'chunked'
    if chunkable:
      serialized = await async_json_dumps(data, sort_keys=True, separators=None, ensure_ascii=True)
      chunkable = len(serialized) > self.CHUNK_SIZE
    #
    if chunkable:
      logger.debug(f"Large packet ({len(serialized)}), chunking...")
      data_hash = str(uuid.uuid4())
      n_chunks = (len(serialized) // ChunkedEmitMixin.CHUNK_SIZE) + int(len(serialized) % ChunkedEmitMixin.CHUNK_SIZE != 0)
      for n in range(0, n_chunks):
        await super().emit(
          'chunked',
          dict(
            id=data_hash, event=evt, chunk=n, total=n_chunks,
            data=serialized[n*self.CHUNK_SIZE:(n+1)*self.CHUNK_SIZE]
          ),
          priority=priority+1,
          **kwargs,
        )
    else:
      await super().emit(evt, data, priority=priority, **kwargs)



import logging
import traceback
from flask import Blueprint, request, redirect, abort, url_for, current_app, jsonify, make_response

from appyter.ext.exceptions import exception_as_dict
from appyter.ext.flask import route_join_with_or_without_slash
from appyter.render.flask_app.constants import get_j2_env
from appyter.render.flask_app.prepare import prepare_data, prepare_request, prepare_results

logger = logging.getLogger(__name__)

core = Blueprint('__main__', __name__)

@route_join_with_or_without_slash(core, methods=['POST'])
def post_index():
  mimetype = request.accept_mimetypes.best_match([
    'text/html',
    'application/json',
  ], 'text/html')
  #
  try:
    data = dict(_config=current_app.config)
    data.update(prepare_data(request))
    data.update(prepare_request(request))
    instance_id = prepare_results(data)
    error = None
  except KeyboardInterrupt:
    raise
  except Exception as e:
    logger.error(traceback.format_exc())
    error = exception_as_dict(e)
  #
  if mimetype in {'text/html'}:
    if error: abort(406)
    else: return redirect(url_for('__main__.data_files', path=instance_id + '/', executor=data.get('_executor'), storage=data.get('_storage')), 303)
  elif mimetype in {'application/json'}:
    if error is not None:
      return make_response(jsonify(error=error), 406)
    else:
      # NOTE: Legacy session_id preserved but deprecated
      ret = dict(_id=instance_id, session_id=instance_id)
      if data.get('_executor'): ret.update(_executor=data['_executor'])
      if data.get('_storage'): ret.update(_storage=data['_storage'])
      return make_response(jsonify(ret), 200)
  else:
    abort(404)

@route_join_with_or_without_slash(core, 'ssr', methods=['POST'])
def post_ssr():
  env = get_j2_env()
  try:
    ctx = request.get_json()
    assert ctx['field'].endswith('Field'), 'Invalid field'
    return env.globals[ctx['field']](**ctx['args']).render()
  except KeyboardInterrupt:
    raise
  except Exception as e:
    return make_response(jsonify(error=exception_as_dict(e)), 406)



import json
import logging
logger = logging.getLogger(__name__)

from appyter.execspec.implementations.wes import WESExecutor
from appyter.ext.contextlib import ensure_context
from appyter.ext.fsspec.core import url_to_fs_ex
from appyter.ext.fsspec.util import fsspec_cp
from appyter.ext.fsspec.sbfs import SBFSFileSystem
from appyter.ext.fsspec.chroot import ChrootFileSystem
from appyter.ext.fsspec.writecache import WriteCacheFileSystem
from appyter.ext.urllib import join_slash, join_url
from appyter.ext.asyncio.helpers import ensure_async, ensure_async_contextmanager


class CavaticaExecutor(WESExecutor):
  ''' Run executions in CAVATICA
  cavatica?project=myuser/myproject&auth_token=MYKEY&cwl=example.cwl
  '''
  protocol = 'cavatica'

  def __init__(self,
    url=None,
    project=None,
    auth_token=None,
    api_endpoint='https://cavatica-api.sbgenomics.com',
    wes_endpoint='wes://cavatica-ga4gh-api.sbgenomics.com',
    drs_endpoint='drs://cavatica-ga4gh-api.sbgenomics.com',
    params={},
    headers={},
    **kwargs,
  ):
    assert project, 'CAVATICA Project required'
    assert auth_token, 'CAVATICA API Key required'
    super().__init__(
      url=wes_endpoint,
      api_endpoint=api_endpoint,
      drs_endpoint=drs_endpoint,
      project=project,
      auth_token=auth_token,
      params=dict(params, workflow_params=dict(params.get('workflow_params', {}), project=project)),
      headers=dict(headers, **{ 'X-SBG-Auth-Token': auth_token }),
      **kwargs,
    )

  async def __aenter__(self):
    await super().__aenter__()
    self.fs = WriteCacheFileSystem(
      fs=ChrootFileSystem(
        fs=SBFSFileSystem(
          auth_token=self.executor_options['auth_token'],
          api_endpoint='https://cavatica-api.sbgenomics.com',
          drs_endpoint='drs://cavatica-ga4gh-api.sbgenomics.com',
        ),
        fo=self.executor_options['project'],
      ),
    )
    self.fs_ctx = ensure_async_contextmanager(self.fs)
    await self.fs_ctx.__aenter__()
    # WES establishes an authenticated client session
    # augment CWL for CAVATICA -- namely, the revision must be specified
    #  we'll use cavatica's app API to determine which revision is necessary
    async with self.session.get(
      join_slash(self.executor_options['api_endpoint'], 'v2', 'apps', self.executor_options['project'], self.cwl['id'])
    ) as req:
      if req.status == 404:
        self.cwl['id'] = f"{self.cwl['id']}/0"
      else:
        res = await req.json()
        if json.dumps(res['raw']['hints'], sort_keys=True) == json.dumps(self.cwl['hints'], sort_keys=True):
          self.cwl['id'] = f"{self.cwl['id']}/{res['revision']}"
        else:
          self.cwl['id'] = f"{self.cwl['id']}/{res['revision']+1}"
    logger.info(f"{self.cwl['id']=}")
    return self

  async def __aexit__(self, type, value, traceback):
    await self.fs_ctx.__aexit__(type, value, traceback)
    await super().__aexit__(type, value, traceback)

  async def _prepare(self, job):
    wes_job = await super()._prepare(job)
    # TODO: could potentially use CAVATICA's DRS import for this
    for k, v in wes_job['workflow_params']['inputs'].items():
      if type(v) == dict and v.get('class') == 'File':
        if not v['path'].startswith(self.executor_options['drs_endpoint']):
          logger.getChild(k).debug(f"identified file path {v['path']}...")
          filename = v['name']
          fs, fspath = url_to_fs_ex(v['path'])
          async with ensure_async(ensure_context(fs)) as fs:
            # attempt to use the filesystem's own drs provider
            logger.getChild(k).debug('trying get_drs...')
            try: drs_uri = await ensure_async(fs.get_drs)(fspath)
            except: drs_uri = None
            # fall back to copying over to CAVATICA-fs
            await fsspec_cp(
              left_fs=fs,
              left_fo=fspath,
              right_fs=self.fs,
              right_fo=f"appyter/input/{job['session']}-{filename}",
            )
            drs_uri = await ensure_async(self.fs.get_drs)(f"appyter/input/{job['session']}-{filename}")
          # use the drs uri
          logger.getChild(k).debug(f"{drs_uri=}")
          v['path'] = drs_uri
    # use a temporary location since CAVATICA doesn't support subdirectories
    wes_job['workflow_params']['inputs']['w'] = f"tmp-appyter-output-{job['session']}"
    return wes_job

  async def _finalize(self, job, wes_job):
    await super()._finalize(job, wes_job)
    # move appyter output to the correct subdirectory
    await ensure_async(self.fs.mv)(
      wes_job['workflow_params']['inputs']['w'],
      f"appyter/output/{job['session']}",
      recursive=True,
    )



'''
A DRS endpoint for accessing `storage://input`
'''

import re
import datetime as dt
import logging
logger = logging.getLogger(__name__)

from flask import current_app, request, send_file, abort
from appyter.render.flask_app.core import core
from appyter.render.flask_app.constants import get_input_fs
from appyter.ext.urllib import url_filename
from appyter import __version__

@core.route('/ga4gh/drs/v1/service-info', methods=['GET'])
def drs_service_info():
  from pathlib import PurePath
  from appyter.ext.re import slugify
  name = slugify(PurePath(current_app.config['IPYNB']).stem.lower())
  return {
    'id': f"cloud.maayanlab.appyter.{name}",
    'name': 'Appyter',
    'type': {
      'group': 'cloud.maayanlab.appyter',
      'artifact': 'drs',
      'version': '1.0.0',
    },
    'organization': {
      'name': 'MaayanLab',
      'url': 'https://maayanlab.cloud',
    },
    'version': __version__,
  }

@core.route('/ga4gh/drs/v1/objects/<string:file_id>', methods=['GET'])
def drs_objects(file_id):
  try:
    fs = get_input_fs()
    file_info = fs.info(file_id)
    assert file_info['type'] == 'file', 'NotAFile'
  except:
    return abort(404)
  #
  res = {
    'id': file_id,
    'self_uri': f"{re.sub(r'^https?://', 'drs://', (current_app.config.get('PUBLIC_URL') or request.url_root).strip('/'))}/{file_id}",
    'access_methods': [
      {
        'access_id': 'primary',
        'type': 'https',
      },
    ],
    'checksums': [
      {
        'type': 'sha-1',
        'checksum': file_id,
      }
    ],
    'created_time': (
      dt.datetime.fromtimestamp(file_info['created'])
      if 'created' in file_info else
      dt.datetime.now()
    ).strftime('%Y-%m-%dT%H:%M:%SZ'),
    'size': file_info['size'],
  }
  return res

@core.route('/ga4gh/drs/v1/objects/<string:file_id>/access/<string:access_id>', methods=['GET'])
def drs_objects_access(file_id, access_id):
  if access_id != 'primary':
    return abort(404)
  #
  try:
    fs = get_input_fs()
    file_info = fs.info(file_id)
    assert file_info['type'] == 'file', 'NotAFile'
  except:
    return abort(404)
  #
  return {
    'url': f"{(current_app.config.get('PUBLIC_URL') or request.url_root).strip('/')}/ga4gh/drs/v1/objects/{file_id}/fetch",
  }

@core.route('/ga4gh/drs/v1/objects/<string:file_id>/fetch', methods=['GET'])
def drs_objects_fetch(file_id):
  try:
    fs = get_input_fs()
    file_info = fs.info(file_id)
    assert file_info['type'] == 'file', 'NotAFile'
  except:
    return abort(404)
  #
  return send_file(fs.open(file_info['name'], 'rb'), attachment_filename=url_filename(file_info['name']), conditional=True)



import sys
import contextlib

@contextlib.contextmanager
def with_many(**withables):
  ctx = {
    name: withable.__enter__()
    for name, withable in withables.items()
  }
  try:
    yield ctx
  except:
    for withable in withables.values():
      withable.__exit__(*sys.exc_info())
  else:
    for withable in withables.values():
      withable.__exit__(None, None, None)

@contextlib.contextmanager
def ensure_context(obj):
  if getattr(obj, '__enter__', None) is not None:
    ctx = obj.__enter__()
  else:
    ctx = obj
  try:
    yield ctx
  except:
    if getattr(obj, '__exit__', None) is not None:
      obj.__exit__(*sys.exc_info())
    else:
      raise
  else:
    if getattr(obj, '__exit__', None) is not None:
      obj.__exit__(None, None, None)

class ContextManagerAsHandle:
  def __init__(self, inner):
    self.inner = inner
  
  def open(self):
    return self.inner.__enter__()
  
  def close(self):
    return self.inner.__exit__(None, None, None)



import time
import asyncio
import contextlib
from appyter.ext.asyncio import helpers
from appyter.ext.itertools import alist

import logging

logger = logging.getLogger(__name__)

import pytest
from appyter.ext.pytest import assert_eq, assert_exc
from appyter.ext.asyncio.event_loop import with_event_loop
@pytest.fixture(scope="session", autouse=True)
def event_loop_fixture():
  with with_event_loop():
    yield

@contextlib.asynccontextmanager
async def asyncctx(a, *, b=1):
  await asyncio.sleep(0.1)
  yield a+b

async def asyncgenfun(a, *, b=1):
  yield a
  await asyncio.sleep(0.1)
  yield a+b

async def asyncfun(a, *, b=1):
  await asyncio.sleep(0.1)
  return a+b

@contextlib.contextmanager
def syncctx(a, *, b=1):
  time.sleep(0.1)
  yield a+b

def syncgenfun(a, *, b=1):
  yield a
  time.sleep(0.1)
  yield a+b

def syncfun(a, *, b=1):
  time.sleep(0.1)
  return a+b

async def async_mixedfun(a, *, b=1):
  return await helpers.ensure_async(syncfun(a, b=b))

def sync_exc():
  logger.info("Sync: Raising Runtime Error")
  raise RuntimeError()

async def async_exc():
  logger.info("Async: Raising Runtime Error")
  raise RuntimeError()

def test_ensure_sync_async_ctx():
  with helpers.ensure_sync(asyncctx(1)) as ctx: assert_eq(ctx, 2)
  with helpers.ensure_sync(asyncctx)(0,b=2) as ctx: assert_eq(ctx, 2)

def test_ensure_sync_async_gen():
  assert_eq(list(helpers.ensure_sync(asyncgenfun)(1)), [1,2])
  assert_eq(list(helpers.ensure_sync(asyncgenfun(0,b=2))), [0,2])

def test_ensure_sync_async_fun():
  assert_eq(helpers.ensure_sync(asyncfun)(1), 2)
  assert_eq(helpers.ensure_sync(asyncfun(0,b=2)), 2)

def test_ensure_sync_async_exc():
  with assert_exc(RuntimeError):
    helpers.ensure_sync(async_exc)()

def test_ensure_sync_sync_ctx():
  i = 0
  with helpers.ensure_sync(syncctx(1)) as ctx:
    assert_eq(ctx, 2)
    i += 1
  with helpers.ensure_sync(syncctx)(0,b=2) as ctx:
    assert_eq(ctx, 2)
    i += 1
  assert_eq(i, 2)

def test_ensure_sync_sync_gen():
  assert_eq(list(helpers.ensure_sync(syncgenfun)(1)), [1,2])
  assert_eq(list(helpers.ensure_sync(syncgenfun(0,b=2))), [0,2])

def test_ensure_sync_sync_fun():
  assert_eq(helpers.ensure_sync(syncfun)(1), 2)
  assert_eq(helpers.ensure_sync(syncfun(0,b=2)), 2)

def test_ensure_sync_sync_exc():
  with assert_exc(RuntimeError):
    helpers.ensure_sync(sync_exc)()

def test_ensure_sync_mixed_fun():
  assert_eq(helpers.ensure_sync(async_mixedfun(0,b=2)), 2)

async def _test_ensure_async_async_ctx():
  i = 0
  async with helpers.ensure_async(asyncctx(1)) as ctx:
    i += 1
    assert_eq(ctx, 2)
  # Not supported
  # async with helpers.ensure_async(asyncctx)(0,b=2) as ctx: assert_eq(ctx, 2)
  assert_eq(i, 1)
def test_ensure_async_async_ctx():
  helpers.ensure_sync(_test_ensure_async_async_ctx)()

async def _test_ensure_async_async_gen():
  assert_eq(await alist(helpers.ensure_async(asyncgenfun)(1)), [1,2])
  assert_eq(await alist(helpers.ensure_async(asyncgenfun(0,b=2))), [0,2])
def test_ensure_async_async_gen():
  helpers.ensure_sync(_test_ensure_async_async_gen)()

async def _test_ensure_async_async_fun():
  assert_eq(await helpers.ensure_async(asyncfun)(1), 2)
  assert_eq(await helpers.ensure_async(asyncfun(0,b=2)), 2)
def test_ensure_async_async_fun():
  helpers.ensure_sync(_test_ensure_async_async_fun)()

async def _test_ensure_async_async_exc():
  with assert_exc(RuntimeError):
    await helpers.ensure_async(async_exc)()
def test_ensure_async_async_exc():
  helpers.ensure_sync(_test_ensure_async_async_exc)()

async def _test_ensure_async_sync_ctx():
  i = 0
  async with helpers.ensure_async(syncctx(1)) as ctx:
    i += 1
    assert_eq(ctx, 2)
  # Not supported
  # async with helpers.ensure_async(syncctx)(0,b=2) as ctx: assert_eq(ctx, 2)
  assert_eq(i, 1)

def test_ensure_async_sync_ctx():
  helpers.ensure_sync(_test_ensure_async_sync_ctx)()

async def _test_ensure_async_sync_gen():
  assert_eq(await alist(helpers.ensure_async(syncgenfun)(1)), [1,2])
  assert_eq(await alist(helpers.ensure_async(syncgenfun(0,b=2))), [0,2])
def test_ensure_async_sync_gen():
  helpers.ensure_sync(_test_ensure_async_sync_gen)()

async def _test_ensure_async_sync_fun():
  assert_eq(await helpers.ensure_async(syncfun)(1), 2)
  assert_eq(await helpers.ensure_async(syncfun(0,b=2)), 2)
def test_ensure_async_sync_fun():
  helpers.ensure_sync(_test_ensure_async_sync_fun)()

async def _test_ensure_async_sync_exc():
  with assert_exc(RuntimeError):
    await helpers.ensure_async(sync_exc)()
def test_ensure_async_sync_exc():
  helpers.ensure_sync(_test_ensure_async_sync_exc)()

async def _test_ensure_async_mixed_fun():
  assert_eq(await helpers.ensure_async(async_mixedfun(0,b=2)), 2)
def test_ensure_async_mixed_fun():
  helpers.ensure_sync(_test_ensure_async_mixed_fun)()

async def _test_async_sync_bottom_up_exc():
  try:
    await helpers.ensure_async(sync_exc)()
  except RuntimeError as e:
    assert isinstance(e, RuntimeError)
    logger.info(f"Async: Received {repr(e)}, Re-raising")
    raise

def test_async_sync_bottom_up_exc():
  with assert_exc(RuntimeError):
    logger.info("Starting async_sync_exc")
    helpers.ensure_sync(_test_async_sync_bottom_up_exc)()
    logger.info("Done.")


@contextlib.asynccontextmanager
async def _test_async_sync_top_down_exc():
  try:
    yield
  except KeyboardInterrupt as e:
    assert isinstance(e, KeyboardInterrupt)
    logger.info(f"Async: Received {repr(e)}, Re-raising")
    raise

def test_async_sync_top_down_exc():
  with assert_exc(KeyboardInterrupt):
    with helpers.ensure_sync(_test_async_sync_top_down_exc()):
      time.sleep(0.5)
      raise KeyboardInterrupt


def test_async_sync_async_ctx():
  class MyAsynCls:
    def __init__(self):
      logger.debug("MyAsynCls()")

    async def __aenter__(self):
      logger.debug("MyAsynCls.__aenter__")
      await asyncio.sleep(0.1)
      return self
    async def __aexit__(self, *args):
      logger.debug("MyAsynCls.__aexit__")
      await asyncio.sleep(0.1)
  class MySynCls:
    def __init__(self):
      logger.debug("MySynCls()")
      self.asyn = MyAsynCls()

    def __enter__(self):
      logger.debug("MySynCls.__enter__")
      helpers.ensure_sync(self.asyn.__aenter__())
      time.sleep(0.1)

    def __exit__(self, *args):
      helpers.ensure_sync(self.asyn.__aexit__(*args))
      time.sleep(0.1)
      logger.debug("MySynCls.__exit__")

  async def _test():
    syn = MySynCls()
    logger.debug("with MySynCls")
    # with syn: # NOTE: this would cause a hang
    async with helpers.ensure_async(syn):
      logger.debug("entered")
      await asyncio.sleep(0.1)
      logger.debug("exited")
  helpers.ensure_sync(_test())



import appyter.ext.multiprocessing

import os
import logging

import pytest
from appyter.ext.pytest import assert_eq
from appyter.ext.asyncio.event_loop import with_event_loop
@pytest.fixture(scope="session", autouse=True)
def event_loop_fixture():
  with with_event_loop():
    yield

FSSPEC_URI = os.environ.get('FSSPEC_URI')

@pytest.mark.skipif(not FSSPEC_URI or 'sbfs://' not in FSSPEC_URI, reason='FSSPEC_URI with sbfs:// necessary to test sbfs')
def test_sbfs():
  import uuid
  import appyter.ext.fsspec
  from pathlib import PurePosixPath
  from appyter.ext.fsspec.core import url_to_fs_ex
  fs, fs_path = url_to_fs_ex(FSSPEC_URI)
  logging.debug(f"{fs=}")
  path = PurePosixPath(fs_path) / f"appyter-test-{str(uuid.uuid4())}"
  logging.debug(path)
  import time
  with fs as fs:
    try:
      logging.debug(fs.ls(''))
      fs.makedirs(fs_path, exist_ok=True)
      logging.debug(f"{fs.ls(fs_path)}")
      fs.makedirs(str(path), exist_ok=True)
      logging.debug(f"{fs.ls(str(path))}")
      with fs.open(str(path/'a'), 'wb') as fw:
        fw.write(b'A')
      with fs.open(str(path/'a'), 'rb') as fr:
        assert_eq(fr.read(), b'A')
      fs.makedirs(str(path/'b'))
      with fs.open(str(path/'b/c'), 'wb') as fw:
        fw.write(b'C')
      with fs.open(str(path/'b/c'), 'rb') as fr:
        assert_eq(fr.read(), b'C')
      with fs.open(str(path/'a'), 'wb') as fw:
        fw.write(b'a')
      assert_eq(fs.cat(str(path/'a')), b'a')
      assert_eq(frozenset(fs.glob(str(path/'**'))), frozenset([str(path/'a'), str(path/'b'), str(path/'b/c')]))
    finally:
      fs.rm(str(path), recursive=True)



from s3fs.core import S3FileSystem, buck_acls, ClientError, translate_boto_error, ParamValidationError, sync_wrapper

class S3FileSystemEx(S3FileSystem):

    async def _info(self, path, bucket=None, key=None, refresh=False, version_id=None):
        if len(path) > 1024: raise FileNotFoundError
        return await super()._info(path, bucket=bucket, key=key, refresh=refresh, version_id=version_id)

    info = sync_wrapper(_info)

    async def _mkdir(self, path, acl="", create_parents=True, **kwargs):
        '''
        If we create an empty directory, we'll add it to `dircache`
        '''
        path = self._strip_protocol(path).rstrip("/")
        if not path:
            raise ValueError
        bucket, key, _ = self.split_path(path)
        if await self._exists(bucket):
            if not key:
                # requested to create bucket, but bucket already exist
                raise FileExistsError
            else:
              try:
                path_info = await self._info(path)
                if path_info['type'] == 'file':
                  raise FileExistsError
              except:
                self.dircache[path] = []
        elif not key or create_parents:
            if acl and acl not in buck_acls:
                raise ValueError("ACL not in %s", buck_acls)
            try:
                params = {"Bucket": bucket, "ACL": acl}
                region_name = kwargs.get("region_name", None) or self.client_kwargs.get(
                    "region_name", None
                )
                if region_name:
                    params["CreateBucketConfiguration"] = {
                        "LocationConstraint": region_name
                    }
                await self._call_s3("create_bucket", **params)
                self.invalidate_cache("")
                self.invalidate_cache(bucket)
            except ClientError as e:
                raise translate_boto_error(e)
            except ParamValidationError as e:
                raise ValueError("Bucket create failed %r: %s" % (bucket, e))
        else:
            # raises if bucket doesn't exist and doesn't get create flag.
            await self._ls(bucket)

    mkdir = sync_wrapper(_mkdir)


''' Async helpers to convert anything sync to and from anything async without deadlocks.
This requires that you initialize the event loop with the appyter.ext.asyncio.event_loop methods
 which will run the event loop in a dedicated thread and ensure all `sync` methods are executed
 on that dedicated event loop.
You only need to use `ensure_async` or `ensure_sync` these methods will deal with dispatching to
 the relevant handlers.
It works with already-sync/async, context managers, generators, & functions of these.
'''
import asyncio
import threading
import functools

from appyter.ext.asyncio.event_loop import get_event_loop

from inspect import (
  isasyncgen,
  isasyncgenfunction,
  isawaitable,
  iscoroutine,
  iscoroutinefunction,
  isfunction,
  isgenerator,
  isgeneratorfunction,
  ismethod,
)

def isasynccontextmanager(obj):
  return getattr(obj, '__aenter__', None) is not None

def iscontextmanager(obj):
  return getattr(obj, '__enter__', None) is not None

def _call(func, args, kwargs):
  return func(*args, **kwargs)

async def ensure_async_literal(obj):
  return obj

async def awaitable_identity(obj):
  return await obj

def ensure_async_func(func):
  @functools.wraps(func)
  async def wrapper(*args, **kwargs):
    return await get_event_loop().run_in_executor(
      None,
      functools.partial(_call, func),
      args,
      kwargs,
    )
  return wrapper

@ensure_async_func
def anext(generator):
  try:
    return next(generator)
  except StopIteration:
    raise StopAsyncIteration

async def ensure_async_generator(generator):
  while True:
    try:
      yield await anext(generator)
    except StopAsyncIteration:
      break

def ensure_sync_coro(coro):
  loop = get_event_loop()
  assert loop._thread_id != threading.current_thread().ident, "Can't run sync coro in async thread!"
  done = threading.Event()
  future = asyncio.run_coroutine_threadsafe(coro, loop)
  future.add_done_callback(lambda *args, **kwargs: done.set())
  try:
    done.wait()
  except KeyboardInterrupt:
    # Cancel the future if this was interrupted and give it time to handle
    #  the cancel exception.
    future.cancel()
    done.wait()
    raise KeyboardInterrupt
  except Exception:
    # Re-raise after cancel without future cancel handling
    future.cancel()
    raise
  return future.result()

def ensure_sync_generator(asyncgenerator):
  __anext__ = ensure_sync_wrapper(asyncgenerator.__anext__)
  while True:
    try:
      value = __anext__()
      yield value
    except StopAsyncIteration:
      break

class _AsyncSyncContextManager:
  def __init__(self, contextmanager):
    self._aenter = contextmanager.__aenter__ if getattr(contextmanager, '__aenter__', None) is not None else ensure_async_func(contextmanager.__enter__)
    self._aexit = contextmanager.__aexit__ if getattr(contextmanager, '__aexit__', None) is not None else ensure_async_func(contextmanager.__exit__)
  async def __aenter__(self):
    return await self._aenter()
  async def __aexit__(self, typ, value, traceback):
    return await self._aexit(typ, value, traceback)

def ensure_async_contextmanager(contextmanager):
  return _AsyncSyncContextManager(contextmanager)

class _SyncAsyncContextManager:
  def __init__(self, asynccontextmanager):
    self._enter = asynccontextmanager.__enter__ if getattr(asynccontextmanager, '__enter__', None) is not None else ensure_sync_wrapper(asynccontextmanager.__aenter__)
    self._exit = asynccontextmanager.__exit__ if getattr(asynccontextmanager, '__exit__', None) is not None else ensure_sync_wrapper(asynccontextmanager.__aexit__)
  def __enter__(self):
    return self._enter()
  def __exit__(self, typ, value, traceback):
    return self._exit(typ, value, traceback)

def ensure_sync_contextmanager(asynccontextmanager):
  return _SyncAsyncContextManager(asynccontextmanager)

def ensure_async_wrapper(callable):
  @functools.wraps(callable)
  def wrapper(*args, **kwargs):
    obj = callable(*args, **kwargs)
    return ensure_async(obj)
  return wrapper

def ensure_async(obj):
  if (
    iscoroutinefunction(obj)
    or isgeneratorfunction(obj)
  ):
    return ensure_async_wrapper(obj)
  elif (
    isasyncgen(obj)
    or isasyncgenfunction(obj)
    or iscoroutine(obj)
    or isawaitable(obj)
    or isasynccontextmanager(obj)
  ):
    return obj
  elif isgenerator(obj):
    return ensure_async_generator(obj)
  elif iscontextmanager(obj):
    return ensure_async_contextmanager(obj)
  elif (
    callable(obj)
    or ismethod(obj)
    or isfunction(obj)
  ):
    return ensure_async_func(obj)
  else:
    return ensure_async_literal(obj)

def ensure_sync_wrapper(callable):
  @functools.wraps(callable)
  def wrapper(*args, **kwargs):
    obj = callable(*args, **kwargs)
    return ensure_sync(obj)
  return wrapper

def ensure_sync(obj):
  if (
    isgeneratorfunction(obj)
    or iscontextmanager(obj)
  ):
    return obj
  elif isasynccontextmanager(obj):
    return ensure_sync_contextmanager(obj)
  elif isasyncgen(obj):
    return ensure_sync_generator(obj)
  elif (
    callable(obj)
    or isfunction(obj)
    or iscoroutinefunction(obj)
  ):
    return ensure_sync_wrapper(obj)
  elif iscoroutine(obj):
    return ensure_sync_coro(obj)
  elif isawaitable(obj):
    return ensure_sync_coro(awaitable_identity(obj))
  else:
    return obj



import logging
from appyter.ext.fsspec.spec.composable import ComposableAbstractFileSystem

from appyter.ext.urllib import join_slash
logger = logging.getLogger(__name__)

from fsspec import AbstractFileSystem
from appyter.ext.fsspec.spec.mountable import MountableAbstractFileSystem
from appyter.ext.fsspec.core import url_to_fs_ex

class MapperFileSystem(MountableAbstractFileSystem, AbstractFileSystem):
  ''' MapperFS is the inverse of a fsspec.mapper -- it lets you use a mapping to
  define a virtual filesystem.
  '''
  root_marker = ''
  protocol = 'mapperfs'

  def __init__(self, pathmap={}, **kwargs):
    '''
    Parameters
    ----------
    pathmap: dict
      A mapping from [path to map onto the target filesystem]: filesystem url to present
    '''
    super().__init__(**kwargs)
    if not pathmap:
      raise ValueError(
          "pathmap is required"
      )
    
    self.pathmap = {}
    for path, fs_uri in pathmap.items():
      if type(fs_uri) == str:
        self.pathmap[path] = url_to_fs_ex(fs_uri)
      elif type(fs_uri) == tuple or type(fs_uri) == list:
        self.pathmap[path] = fs_uri
      else:
        self.pathmap[path] = fs_uri, ''

    self.listing = {}
    for mapping in self.pathmap:
      src_split = mapping.split('/')
      for i in range(len(src_split)):
        p = '/'.join(src_split[:i+1])
        if i == 0:
          if '' not in self.listing:
            self.listing[''] = {}
          if p:
            self.listing[''][p] = True
        else:
          self.listing['/'.join(src_split[:i])][p] = True
        if i < len(src_split)-1:
          if p not in self.listing:
            self.listing[p] = {}

  def _pathmap(self, path):
    ''' Return (fs, path) depending on whether we hit a mapped paths or not
    '''
    if path in self.pathmap:
      return (*self.pathmap[path], path)
    else:
      path_split = path.split('/')
      for i in range(1, len(path_split)):
        base_path = '/'.join(path_split[:-i])
        if base_path in self.pathmap:
          fs, fo = self.pathmap[base_path]
          return fs, join_slash(fo, *path_split[-i:]), base_path
      raise FileNotFoundError(path)

  def __enter__(self):
    for fs, _fo in self.pathmap.values():
      if getattr(fs, '__enter__', None) is not None:
        fs.__enter__()
    return self

  def __exit__(self, type, value, traceback):
    for fs, _fo in self.pathmap.values():
      if getattr(fs, '__exit__', None) is not None:
        fs.__exit__(type, value, traceback)

  def mkdir(self, path, **kwargs):
    raise PermissionError(path)
  
  def makedirs(self, path, exist_ok=False):
    raise PermissionError(path)

  def rmdir(self, path):
    raise PermissionError(path)

  def rm_file(self, path):
    raise PermissionError(path)

  def rm(self, path, recursive=False, maxdepth=None):
    raise PermissionError(path)

  def cat_file(self, path, start=None, end=None, **kwargs):
    fs, fs_path, _ = self._pathmap(path)
    return fs.cat_file(fs_path, start=start, end=end, **kwargs)

  def copy(self, path1, path2, recursive=False, on_error=None, maxdepth=None, **kwargs):
    raise PermissionError(path2)

  def mv(self, path1, path2, recursive=False, maxdepth=None, **kwargs):
    raise PermissionError(path1)

  def exists(self, path, **kwargs):
    return path in self.listing or path in self.pathmap

  def get_drs(self, path, **kwargs):
    if path in self.listing:
      raise IsADirectoryError
    else:
      fs, fs_path, _ = self._pathmap(path)
      return fs.get_drs(fs_path, **kwargs)

  def info(self, path, **kwargs):
    if path in self.listing:
      return {
        'name': path,
        'type': 'directory',
      }
    else:
      fs, fs_path, _ = self._pathmap(path)
      info = fs.info(fs_path, **kwargs)
      info = dict(info, name=path)
      return info

  def ls(self, path, detail=False, **kwargs):
    ''' Aggregate results based on pathmap listing & underlying fs
    '''
    logger.debug(f"ls({path})")
    results = []
    if path in self.listing:
      for p in self.listing[path]:
        if detail:
          results.append(self.info(p))
        else:
          results.append(p)
    else:
      try:
        fs, fo, base_path = self._pathmap(path)
        for p in fs.ls(fo, detail=detail):
          if detail:
            results.append(dict(p, name=join_slash(base_path, p['name'])))
          else:
            results.append(join_slash(base_path, p))
      except FileNotFoundError:
        pass
    return results

  def _open(self, path, mode="rb", block_size=None, autocommit=True, cache_options=None, **kwargs):
    fs, fs_path, _ = self._pathmap(path)
    if 'r' not in mode: raise PermissionError(path)
    return fs._open(
      fs_path,
      mode=mode,
      block_size=block_size,
      autocommit=autocommit,
      cache_options=cache_options,
      **kwargs,
    )

  def to_json(self):
    import json
    cls = type(self)
    cls = '.'.join((cls.__module__, cls.__name__))
    proto = (
      self.protocol[0]
      if isinstance(self.protocol, (tuple, list))
      else self.protocol
    )
    return json.dumps(
      dict(
        **{"cls": cls, "protocol": proto, "args": self.storage_args},
        **dict(
          self.storage_options,
          pathmap={
            path: [json.loads(fs.to_json()), fo]
            for path, (fs, fo) in self.pathmap.items()
          },
        )
      )
    )

  @staticmethod
  def from_json(blob):
    import json
    from fsspec.registry import _import_class, get_filesystem_class
    dic = json.loads(blob)
    protocol = dic.pop("protocol")
    try:
      cls = _import_class(dic.pop("cls"))
    except (ImportError, ValueError, RuntimeError, KeyError, AttributeError):
      cls = get_filesystem_class(protocol)
    if cls.from_json == MapperFileSystem.from_json:
      return cls(
        *dic.pop("args", ()),
        pathmap={
          path: (ComposableAbstractFileSystem.from_json(json.dumps(fs)), fo)
          for path, [fs, fo] in dic.pop('pathmap').items()
        },
        **dic,
      )
    else:
      return cls.from_json(blob)



import traceback
import logging
import shutil
from fsspec.core import url_to_fs
from flask import request, jsonify, abort
from appyter.ext.asyncio.helpers import ensure_async, ensure_sync
from appyter.ext.contextlib import ContextManagerAsHandle

from appyter.ext.fsspec.core import url_to_chroot_fs
from appyter.ext.urllib import URI
from appyter.render.flask_app.constants import get_input_fs
logger = logging.getLogger(__name__)

from appyter.render.flask_app.core import core
from appyter.render.flask_app.socketio import socketio
from appyter.render.flask_app.prepare import prepare_storage, prepare_request
from appyter.ext.flask import secure_filepath, route_join_with_or_without_slash
from appyter.ext.hashlib import sha1sum_io
from appyter.ext.uuid import generate_uuid

# organize file by content hash
def organize_file_content(data_fs, tmp_fs, tmp_path, filename=None):
  with tmp_fs.open(tmp_path, 'rb') as fr:
    content_hash = sha1sum_io(fr)
  if not data_fs.exists(content_hash):
    with tmp_fs.open(tmp_path, 'rb') as fr:
      with data_fs.open(content_hash, 'wb') as fw:
        shutil.copyfileobj(fr, fw)
  return f"storage://input/{content_hash}{('#' + filename) if filename else ''}"

@route_join_with_or_without_slash(core, 'check', '<path:path>', methods=['GET'])
def check(path):
  qs = request.query_string.decode()
  url = path + (('?'+qs) if qs else '')
  fs, fs_path = url_to_fs(url)
  if fs.exists(fs_path):
    return jsonify(fs.info(fs_path))
  else:
    abort(404)

# upload from client
@socketio.on("siofu_start")
async def siofu_start(sid, data):
  try:
    path = generate_uuid()
    filename = secure_filepath(data.get('name'))
    tmp_fs_ctx = ContextManagerAsHandle(url_to_chroot_fs('memory:///'))
    tmp_fs = tmp_fs_ctx.open()
    async with socketio.session(sid) as sess:
      sess['file_%d' % (data.get('id'))] = dict(
        data,
        path=path,
        name=filename,
        bytesLoaded=0,
        tmp_fs_ctx=tmp_fs_ctx,
        tmp_fs=tmp_fs,
        fh=tmp_fs.open(path, 'wb'),
      )
    await socketio.emit(
      'siofu_ready',
      {
        'id': data.get('id'),
        'name': None,
      },
      to=sid,
    )
  except Exception as e:
    logger.error(traceback.format_exc())
    await socketio.emit('siofu_error', str(e), to=sid)

@socketio.on("siofu_progress")
async def siofu_progress(sid, evt):
  async with socketio.session(sid) as sess:
    sess['file_%d' % (evt['id'])]['fh'].write(evt['content'])
  logger.debug(f"progress: {evt}")
  await socketio.emit("siofu_chunk", dict(
    id=evt['id'],
  ), to=sid)

@socketio.on("siofu_done")
async def siofu_done(sid, evt):
  async with socketio.session(sid) as sess:
    fid = 'file_%d' % (evt['id'])
    file = sess[fid]
    file['fh'].close()
    input_fs = get_input_fs()
    tmp_fs_ctx = file['tmp_fs_ctx']
    file_uri = await ensure_async(organize_file_content)(input_fs, file['tmp_fs'], file['path'], file['name'])
    tmp_fs_ctx.close()
    #
    if 'catalog-integration' in sess['config']['EXTRAS']:
      # if you upload a file in a request, it should get registered
      try:
        from appyter.extras.catalog_integration.uploads import FileInfo, add_file
        file_uri_parsed = URI(file_uri)
        await add_file(
          FileInfo(
            file=str(file_uri_parsed.with_fragment(None)),
            filename=file_uri_parsed.fragment,
            metadata=dict(
              size=file['size'],
            ),
          ),
          auth=file.get('meta', {}).get('auth'),
          config=sess['config'],
        )
      except:
        logger.warning(traceback.format_exc())
    #
    del sess[fid]
  #
  await socketio.emit('siofu_complete', dict(
    id=evt['id'],
    detail=dict(full_filename=file_uri)
  ), to=sid)

# upload from client with POST
def upload_from_request(req, fname, multiple=False):
  ret = []
  for fh in (req.files.getlist(fname) if multiple else [req.files.get(fname)]):
    if not fh:
      ret.append(None)
      continue
    filename = secure_filepath(fh.filename)
    if not filename:
      ret.append(None)
      continue
    data = prepare_request(req)
    with url_to_chroot_fs(str(URI(prepare_storage(data)).join('input'))) as input_fs:
      path = generate_uuid()
      with url_to_chroot_fs('memory:///') as tmp_fs:
        with tmp_fs.open(path, 'wb') as fw:
          fh.save(fw)
        file_uri = organize_file_content(input_fs, tmp_fs, path, filename)
    #
    if 'catalog-integration' in data['_config']['EXTRAS']:
      # if you upload a file in a request, it should get registered
      try:
        from appyter.extras.catalog_integration.uploads import FileInfo, add_file
        file_uri_parsed = URI(file_uri)
        ensure_sync(add_file(
          FileInfo(file=str(file_uri_parsed.with_fragment(None)), filename=file_uri_parsed.fragment),
          auth=data.get('_auth'),
          config=data.get('_config'),
        ))
      except:
        logger.warning(traceback.format_exc())
    #
    ret.append(file_uri)
  #
  return ret if multiple else ret[0]



from appyter.ext.asyncio.helpers import ensure_async, ensure_async_contextmanager
from appyter.ext.fsspec.core import url_to_chroot_fs
from appyter.ext.urllib import URI
from appyter.extras.catalog_integration.uploads import list_files
from appyter.extras.catalog_integration.user_config import get_user_config
from appyter.render.flask_app.upload import organize_file_content

async def create_userfs(data):
  from appyter.ext.fsspec.mapperfs import MapperFileSystem
  pathmap = {}
  user_config = await get_user_config(data['_auth'], data['_config'])
  try:
    files = await list_files(data['_auth'], data['_config'])
    pathmap.update({
      f"uploads/{file['filename']}": file['file']['id']
      for file in files
    })
  except:
    pass
  try:
    import re
    if not user_config.get('cavatica_api_key'): raise PermissionError
    assert re.match(r'^[0-9a-f]+$', user_config['cavatica_api_key']) is not None, 'Malformed CAVATICA API Key'
    pathmap['integrations/cavatica'] = f"chroot::sbfs://#?sbfs.auth_token={user_config['cavatica_api_key']}"
  except:
    pass
  return MapperFileSystem(pathmap=pathmap)

async def upload_user_to_storage(storage, data):
  ''' Convert all `user://` file_uris into `drs://` or `storage://` uris
  '''
  # TODO: this is a hack for now,
  #  the whole thing should probably be part of the file field
  files = {
    k[len('_file:'):]: v
    for k, v in data.items()
    if k.startswith('_file:')
  }
  if any(
    file_uri.startswith('user://')
    for file_uri in files.values()
  ):
    from appyter.ext.dict import dict_typed_flatten, dict_typed_unflatten
    flat_data = dict_typed_flatten(data)
    async with ensure_async_contextmanager(await create_userfs(data)) as userfs:
      async with ensure_async_contextmanager(url_to_chroot_fs(str(URI(storage).join('input')))) as storagefs:
        for filename, file_uri in files.items():
          if file_uri.startswith('user://'):
            existing_file_uri = str(URI(file_uri).with_fragment_path(filename))
            try:
              # Use DRS if it's available
              new_file_uri_without_filename = await ensure_async(userfs.get_drs)(file_uri[len('user://'):])
              new_file_uri = str(URI(new_file_uri_without_filename).with_fragment_path(filename))
            except:
              # Otherwise put it into `storage://`
              new_file_uri = await ensure_async(organize_file_content)(storagefs, userfs, file_uri[len('user://'):], filename=filename)
              new_file_uri_without_filename = str(URI(new_file_uri).with_fragment_path(None))
            # replace the uri if it appears in `data`
            flat_data = [
              (K, new_file_uri_without_filename) if K[-1] == f"_file:{filename}" else (K, new_file_uri if v == existing_file_uri else v)
              for K, v in flat_data
            ]
    return dict_typed_unflatten(flat_data)
  return data



def importdir(_dirname_, _package_, _globals_):
  import pathlib, importlib
  _dirname_ = pathlib.Path(_dirname_)
  for f in _dirname_.glob('[!_.]*'):
    if f.is_file() and f.suffix == '.py':
      modname = f.stem
    elif f.is_dir():
      modname = f.name
    else:
      continue
    mod = importlib.import_module('.{}'.format(modname), _package_)
    _globals_.update(**{modname: mod})

def importdir_deep(_dirname_, _package_, _globals_, filter_mod=lambda m, k, v: not k.startswith('_')):
  import pathlib, importlib
  _dirname_ = pathlib.Path(_dirname_)
  for f in _dirname_.glob('[!_.]*'):
    if f.is_file() and f.suffix == '.py':
      modname = f.stem
    elif f.is_dir():
      modname = f.name
    else:
      continue
    mod = importlib.import_module('.{}'.format(modname), _package_)
    _globals_.update(**{
      k: v
      for k, v in mod.__dict__.items()
      if filter_mod(mod, k, v)
    })



import os
import signal
import multiprocessing as mp

def interrupt(proc: mp.Process):
  try:
    os.kill(proc.pid, signal.SIGINT)
  except ProcessLookupError:
    pass
  except KeyboardInterrupt:
    raise
  except:
    proc.terminate()



import os
import asyncio

COPY_BUFSIZE = 1024 * 1024 if os.name == 'posix' else 64 * 1024

async def async_copyfileobj(fsrc, fdst, length=0):
  ''' Async version of shutil.copyfileobj
  '''
  if not length:
    length = COPY_BUFSIZE
  fsrc_read = fsrc.read
  fdst_write = fdst.write
  while True:
    buf = fsrc_read(length)
    if not buf:
      break
    await asyncio.sleep(0)
    fdst_write(buf)
    await asyncio.sleep(0)



import logging
import asyncio
import traceback
import functools
import threading
import contextlib
from concurrent.futures import ThreadPoolExecutor
logger = logging.getLogger(__name__)

class SharedEventLoopThreadPoolExecutor(ThreadPoolExecutor):
  ''' This thread pool will initialize each thread with the same event loop.
  '''
  def __init__(self, *args, loop=None, initializer=None, **kwargs):
    initializer = functools.partial(self._initializer, initializer=initializer)
    super().__init__(*args, initializer=initializer, **kwargs)
    self.loop = loop or asyncio.get_event_loop()
  
  def _initializer(self, *args, initializer=None, **kwargs):
    logger.debug(f"Thread initializer {self.loop=}")
    asyncio.set_event_loop(self.loop)
    if initializer is not None:
      return initializer(*args, **kwargs)

class EventLoopThread(threading.Thread):
  ''' Keep the event loop running in this thread until `stop` is called.
  '''
  def __init__(self, loop=None):
    super().__init__()
    self.close = threading.Event()
    self.loop = loop or asyncio.get_event_loop()

  async def _run(self):
    logger.debug("Event loop started")
    try:
      await self.loop.run_in_executor(None, self.close.wait)
    except:
      logger.error(f"{traceback.format_exc()}")
    finally:
      logger.debug(f"Cancelling active tasks...")
      # we omit our own task since that would cause a deadlock
      #  and we cancel tasks in reverse which works well in python3.10,
      #  typically cancelling tasks launched last first
      tasks = reversed(list(asyncio.all_tasks() - {asyncio.current_task()}))
      for task in tasks:
        task.cancel()
        try:
          await task
        except asyncio.CancelledError:
          pass
        except:
          logger.error(traceback.format_exc())

  def run(self):
    try:
      self.loop.run_until_complete(self._run())
    finally:
      logger.debug(f"Event loop closed.")

  def stop(self):
    logger.debug("Stopping event loop...")
    self.close.set()
    self.join()

_LOOP = None

def new_event_loop():
  ''' Construct a new event loop with some custom extensions:
  - a dedicated thread to run the event loop
  - a thread pool executor that sets the event loop in spawned threads
  '''
  global _LOOP
  if _LOOP is not None:
    logger.warning('Event loop already exists')
    return _LOOP
  logger.debug(f"New event loop")
  try:
    import uvloop
    loop = uvloop.new_event_loop()
  except ImportError:
    logger.warning('Install uvloop for better performance')
    loop = asyncio.new_event_loop()
  asyncio.set_event_loop(loop)
  loop_executor = SharedEventLoopThreadPoolExecutor(loop=loop)
  loop.set_default_executor(loop_executor)
  loop_thread = EventLoopThread(loop=loop)
  loop_thread.start()
  _loop_close = loop.close
  def loop_close(*args, **kwargs):
    global _LOOP
    logger.debug("Closing event loop...")
    loop_thread.stop()
    _LOOP = None
    return _loop_close(*args, **kwargs)
  setattr(loop, 'close', loop_close)
  setattr(loop, '_thread_id', loop_thread.ident)
  _LOOP = loop
  return _LOOP

def get_event_loop():
  ''' Get the same loop in different threads, create the loop if it doesn't exist
  '''
  global _LOOP
  if _LOOP is not None:
    return _LOOP
  else:
    return asyncio.get_running_loop()

@contextlib.contextmanager
def with_event_loop():
  loop = new_event_loop()
  try:
    yield loop
  finally:
    loop.close()



import appyter.ext.multiprocessing

import os
import uuid
import logging

import pytest
from appyter.ext.pytest import assert_eq
from appyter.ext.asyncio.event_loop import with_event_loop
@pytest.fixture(scope="session", autouse=True)
def event_loop_fixture():
  with with_event_loop():
    yield

FSSPEC_URI = os.environ.get('FSSPEC_URI', f"file:///tmp/{str(uuid.uuid4())}")

@pytest.mark.skipif('memory://' in FSSPEC_URI, reason='in-memory store pre-init wont work with fuse in separate process')
def test_fuse_mount():
  from appyter.ext.fsspec.core import url_to_chroot_fs
  fs = url_to_chroot_fs(FSSPEC_URI)
  with fs as fs:
    try:
      fs.makedirs('', exist_ok=True)
      fs.pipe('test', b'Hello World!')
      with fs.mount(fuse=True) as mnt_dir:
        assert mnt_dir.is_mount(), 'Expected mount, got directory'
        logging.debug(f"{mnt_dir=}")
        assert_eq(
          frozenset({
            (p.name, p.open('rb').read())
            for p in mnt_dir.rglob('*') if p.is_file()
          }),
          frozenset({('test', b'Hello World!')}),
        )
        (mnt_dir/'hi').mkdir(exist_ok=True)
        (mnt_dir/'hi/world').open('w').write('test')
        assert_eq((mnt_dir/'hi/world').open('r').read(), 'test')
        # seek works
        with (mnt_dir/'bye').open('w+') as fw:
          fw.write('hi!')
          fw.seek(0)
          fw.write('H')
        assert_eq((mnt_dir/'bye').open('r').read(), 'Hi!')
    finally:
      fs.rm('', recursive=True)

def test_fuseless_mount():
  from appyter.ext.fsspec.core import url_to_chroot_fs
  fs = url_to_chroot_fs(FSSPEC_URI)
  with fs as fs:
    try:
      fs.makedirs('', exist_ok=True)
      fs.pipe('test', b'Hello World!')
      with fs.mount(fuse=False) as mnt_dir:
        assert not mnt_dir.is_mount(), 'Expected directory, got mount'
        logging.debug(f"{mnt_dir=}")
        assert_eq(
          frozenset({
            (p.name, p.open('rb').read())
            for p in mnt_dir.rglob('*') if p.is_file()
          }),
          frozenset({('test', b'Hello World!')}),
        )
    finally:
      fs.rm('', recursive=True)



# inspired by from https://github.com/fsspec/filesystem_spec

_registry = {}
def register_implementation(name, cls):
  global _registry
  _registry[name] = cls

known_implementations = {
  'dispatch': {'class': 'appyter.execspec.implementations.dispatch.DispatchExecutor'},
  'docker': {'class': 'appyter.execspec.implementations.docker.DockerExecutor'},
  'kube': {'class': 'appyter.execspec.implementations.kube.KubernetesExecutor'},
  'local': {'class': 'appyter.execspec.implementations.local.LocalExecutor'},
  'subprocess': {'class': 'appyter.execspec.implementations.subprocess.SubprocessExecutor'},
  'wes': {'class': 'appyter.execspec.implementations.wes.WESExecutor'},
  'cavatica': {'class': 'appyter.execspec.implementations.cavatica.CavaticaExecutor'},
}

def get_executor_class(protocol):
  global _registry
  if protocol not in _registry:
    if protocol not in known_implementations:
      raise ValueError(f"Protocol not known: {protocol}")
    bit = known_implementations[protocol]
    try:
      register_implementation(protocol, _import_class(bit['class']))
    except ImportError as e:
      raise ImportError(bit['err']) from e
  cls = _registry[protocol]
  return cls

def _import_class(cls, minv=None):
    """Take a string FQP and return the imported class or identifier
    clas is of the form "package.module.klass" or "package.module:subobject.klass"
    """
    import importlib
    if ":" in cls:
      mod, name = cls.rsplit(":", 1)
      mod = importlib.import_module(mod)
      for part in name.split("."):
        mod = getattr(mod, part)
      return mod
    else:
      mod, name = cls.rsplit(".", 1)
      mod = importlib.import_module(mod)
      return getattr(mod, name)

def executor(protocol, **executor_options):
  cls = get_executor_class(protocol)
  return cls(**executor_options)



import appyter.ext.multiprocessing

import tempfile
import contextlib
from pathlib import Path

import appyter.ext.fsspec
from appyter.ext.fsspec.core import url_to_chroot_fs
from appyter.ext.fsspec.fuse import fs_mount
from appyter.ext.asyncio.helpers import ensure_sync

import pytest
from appyter.ext.pytest import assert_eq
from appyter.ext.asyncio.event_loop import with_event_loop
@pytest.fixture(scope="session", autouse=True)
def event_loop_fixture():
  with with_event_loop():
    yield

@contextlib.contextmanager
def _test_ctx():
  with tempfile.TemporaryDirectory() as tmpdir:
    tmpdir = Path(tmpdir)
    (tmpdir/'a'/'b').mkdir(parents=True)
    with (tmpdir/'a'/'b'/'c').open('w') as fw:
      fw.write('C')
    with (tmpdir/'a'/'d').open('w') as fw:
      fw.write('D')
    with (tmpdir/'e').open('w') as fw:
      fw.write('E')
    yield tmpdir

def test_file_chroot_fuse():
  with _test_ctx() as tmpdir:
    fs = url_to_chroot_fs(str(tmpdir), pathmap={'g': str(tmpdir/'e')}, cached=True)
    with ensure_sync(fs_mount(fs)) as fs:
      # reads work
      assert_eq(frozenset(str(p.relative_to(fs)) for p in fs.rglob('*')), frozenset(['a', 'a/b', 'a/b/c', 'a/d', 'e', 'g']))
      assert_eq((fs/'a'/'b'/'c').open('rb').read(), b'C')
      assert_eq((fs/'a'/'d').open('rb').read(), b'D')
      assert_eq((fs/'e').open('rb').read(), b'E')
      # pathmap works
      assert_eq((fs/'g').open('rb').read(), b'E')
      # writes work
      with (fs / 'f').open('wb') as fw:
        fw.write(b'f')
      assert_eq((fs/'f').open('rb').read(), b'f')
      # overwrites work
      with (fs / 'f').open('wb') as fw:
        fw.write(b'F')
      assert_eq((fs/'f').open('rb').read(), b'F')
      # and make it to the underlying store
      assert_eq((tmpdir/'f').open('rb').read(), b'F')



import re
from appyter.execspec.registry import get_executor_class

def url_to_executor(url, **kwargs):
  m = re.match(r"^([^:#?]+)(::|://)?(.*)$", url)
  assert m is not None, "Invalid executor spec"
  proto = m.group(1)
  if m.group(2) == '::':
    url = m.group(3)
  else:
    url = m.group(0)
  cls = get_executor_class(proto)
  return cls(**cls.parse(url), **kwargs)



import os
import json
import click
import fsspec

from appyter.context import get_env, get_jinja2_env
from appyter.ext.click import click_option_setenv, click_argument_setenv
from appyter.ext.urllib import join_slash
from appyter.parse.nb import nb_from_ipynb_io
from appyter.render.nbinspect.cli import nbinspect
from appyter.render.nbinspect.jsonschema import render_jsonschema_from_nbtemplate

def render_openapi_from_nbtemplate(env, nb):
  ''' Render a openapi representing the relevant Fields throughout the notebook.
  TODO: outputs
  TODO: appyter.json metadata
  '''
  schema = {
    'openapi': '3.1.0',
    'info': {
      # TODO
      'version': '1.0.0',
      'title': 'Appyter API',
      'description': 'An API for interfacing with the Appyter.',
      # termsOfService:
      'contact': {
        'name': "Ma'ayan Lab",
        # 'email': '',
        'url': 'https://maayanlab.cloud/',
      },
      'license': {
        'name': 'CC-BY-NC-SA',
        'url': 'https://github.com/MaayanLab/appyter/blob/main/LICENSE',
      },
    },
    'servers': [
      { 'url': '/' },
    ],
    'paths': {
      '/': {
        'get': {
          'description': 'Get information about this appyter',
          'operationId': 'appyter_inspect',
          'responses': {
            '200': {
              'description': 'Appyter instance',
              'content': {
                'application/vnd.jupyter': {
                  # 'description': 'Appyter notebook',
                  'schema': {
                    '$ref': '#/components/schemas/AppyterIPYNB',
                  },
                },
                'application/json': {
                  # 'description': 'Appyter nbinspect',
                  'schema': {
                    '$ref': '#/components/schemas/AppyterNBInspect',
                  },
                },
                'text/html': {
                  # 'description': 'Appyter form',
                  'schema': {
                    'type': 'string',
                  },
                },
              },
            },
          },
        },
        'post': {
          'description': 'Executes the appyter',
          'operationId': 'appyter_execute',
          'requestBody': {
            'required': True,
            'content': {
              'multipart/form-data': {
                'schema': render_jsonschema_from_nbtemplate(env, nb),
              },
            },
          },
          'responses': {
            '200': {
              'description': 'Appyter results',
              'content': {
                'application/json': {
                  'schema': {
                    'type': 'object',
                    'properties': {
                      'session_id': {
                        'type': 'string',
                        'description': 'DEPRECATED: A unique identifier for this appyter invokation',
                      },
                      '_id': {
                        'type': 'string',
                        'description': 'A unique identifier for this appyter invokation',
                      },
                      '_storage': {
                        'type': 'string',
                        'description': 'A URI for the storage backend where this appyter is saved when using integrations',
                      },
                    },
                    'required': ['_id']
                  },
                },
              },
            },
            '406': {
              'description': 'Not acceptable error, field(s) did not satisfy constraint(s)',
              'content': {
                'application/json': {
                  'schema': {
                    '$ref': '#/components/schemas/Error',
                  },
                },
              },
            },
          },
        },
      },
      '/{_id}/': {
        'get': {
          'description': 'Fetch or interact with an appyter instance',
          'operationId': 'appyter_instance_fetch',
          'parameters': [
            {
              'in': 'path',
              'name': '_id',
              'schema': {
                'type': 'string'
              },
              'required': True,
              'description': 'Appyter instance identifier',
            },
          ],
          'responses': {
            '200': {
              'description': 'Appyter instance',
              'content': {
                'application/vnd.jupyter': {
                  # 'description': 'Appyter instance',
                  'schema': {
                    "$ref": "#/components/schemas/AppyterInstanceIPYNB",
                  },
                },
                'application/json': {
                  # 'description': 'Appyter instance',
                  'schema': {
                    "$ref": "#/components/schemas/AppyterInstanceIPYNB",
                  },
                },
                'text/html': {
                  # 'description': 'Interactive appyter instance viewer',
                  'schema': {
                    'type': 'string',
                  },
                },
              },
            },
            '404': {
              'description': 'Appyter instance not found',
            },
          },
        },
      },
    },
    'components': {
      'schemas': {
        'AppyterNBInspect': {
          'description': 'Appyter field extracts',
          'type': 'array',
          'items': {
            'type': 'object',
            'properties': {
              'field': {
                'type': 'string',
              },
              'args': {
                'description': 'Field-specific arguments',
                'type': 'object',
                'properties': {
                  'name': {
                    'description': 'The unique name of this field',
                    'type': 'string',
                  },
                  'label': {
                    'description': 'A human readable label for the field',
                    'type': 'string',
                  },
                  'description': {
                    'description': 'A human readable description for the field',
                    'type': 'string',
                  },
                  'choices': {
                    'description': 'A list or lookup dictionary of possible choices',
                    'oneOf': [
                      {'type': 'array', 'items': { '$ref': '#/components/schemas/Any' }},
                      {'type': 'object', 'additionalProperties': { '$ref': '#/components/schemas/Any' }},
                    ]
                  },
                  'required': {
                    'description': 'Whether or not this field is required',
                    'type': 'boolean',
                    'default': False,
                  },
                  'default': {
                    'description': 'The value used if not provided',
                    '$ref': '#/components/schemas/Any',
                  },
                },
                'required': ['name'],
                'additionalProperties': True,
              },
            },
          },
        },
        'AppyterIPYNB': {
          'description': 'Appyter IPython Notebook format',
          'type': 'object',
          'properties': {
            'metadata': {
              'description': 'Jupyter notebook metadata',
              'type': 'object',
              'properties': {
                'appyter': {
                  'description': 'Appyter specific metadata',
                  'type': 'object',
                  'properties': {
                    'info': {
                      '$ref': 'https://raw.githubusercontent.com/MaayanLab/appyter-catalog/master/schema/appyter-validator.json',
                    },
                  },
                },
              },
            },
          },
        },
        'AppyterInstanceIPYNB': {
          'description': 'Appyter Instance IPython Notebook format',
          'type': 'object',
          'properties': {
            'metadata': {
              'description': 'Jupyter notebook metadata',
              'type': 'object',
              'properties': {
                'appyter': {
                  'description': 'Appyter specific metadata',
                  'type': 'object',
                  'properties': {
                    'info': {
                      '$ref': 'https://raw.githubusercontent.com/MaayanLab/appyter-catalog/master/schema/appyter-validator.json',
                    },
                    'nbconstruct': {
                      'type': 'object',
                      'properties': {
                        'filename': {
                          'type': 'string',
                        },
                        'files': {
                          'description': 'File name to source mapping used during nbconstruct',
                          'type': 'object',
                          'additionalProperties': {
                            'type': 'string',
                          },
                        },
                        'version': {
                          'description': 'Version of appyter used for nbconstruct',
                          'type': 'string',
                        },
                      },
                    },
                    'nbexecute': {
                      'type': 'object',
                      'properties': {
                        'started': {
                          'description': 'Server time when nbexecute was started',
                          'type': 'string',
                          'format': 'date-time',
                        },
                        'completed': {
                          'description': 'Server time when nbexecute completed',
                          'type': 'string',
                          'format': 'date-time',
                        },
                        'files': {
                          'description': 'File name to source mapping present after nbexecute',
                          'type': 'object',
                          'additionalProperties': {
                            'type': 'string',
                          },
                        },
                        'version': {
                          'description': 'Version of appyter used for nbexecute',
                          'type': 'string',
                        },
                      },
                    },
                  },
                },
              },
              'additionalProperties': True,
            },
          },
          'additionalProperties': True,
        },
        'Error': {
          'type': 'object',
          'properties': {
            'error': {
              'type': 'object',
              'properties': {
                'cls': {
                  'type': 'string',
                  'description': 'Specific type of error',
                },
                'message': {
                  'type': 'string',
                  'description': 'Human readable indication of what went wrong',
                },
              },
            },
          },
          'additionalProperties': True,
          'required': ['cls'],
        },
        'Any': {
          'oneOf': [
            {'type': 'null'},
            {'type': 'string'},
            {'type': 'number'},
            {'type': 'integer'},
            {'type': 'boolean'},
            {'type': 'array', 'items': { '$ref': '#/components/schemas/Any' } },
            {'type': 'object', 'additionalProperties': { '$ref': '#/components/schemas/Any' } },
          ],
        },
      },
    },
  }
  return schema

@nbinspect.command(help='Create OpenAPI spec for appyter')
@click.option('-o', '--output', default='-', type=click.File('w'), help='The output location of the inspection json')
@click_option_setenv('--cwd', envvar='APPYTER_CWD', default=os.getcwd(), help='The directory to treat as the current working directory for templates and execution')
@click_argument_setenv('ipynb', envvar='APPYTER_IPYNB')
def openapi(cwd, ipynb, output, **kwargs):
  cwd = os.path.realpath(cwd)
  env = get_jinja2_env(config=get_env(cwd=cwd, ipynb=ipynb, mode='inspect', **kwargs))
  with fsspec.open(join_slash(cwd, ipynb), 'r') as fr:
    nbtemplate = nb_from_ipynb_io(fr)
  json.dump(render_openapi_from_nbtemplate(env, nbtemplate), output)



import typing as t
import pathlib
import contextlib
from fsspec import AbstractFileSystem

import logging
logger = logging.getLogger(__name__)

class MountableAbstractFileSystem:
  @contextlib.contextmanager
  def mount(self: AbstractFileSystem, path='', passthrough=False, **kwargs):
    ''' mount fallback: defer to chroot-filesystem
    '''
    from appyter.ext.fsspec.chroot import ChrootFileSystem
    fs = self if isinstance(self, ChrootFileSystem) else ChrootFileSystem(fs=self, fo=path)
    with fs.mount(path=path, passthrough=False, **kwargs) as mount_dir:
      yield mount_dir



from appyter.ext.asyncio.helpers import ensure_sync

class SyncAsyncFileSystem:
  ''' A helper for fsspec AsyncFileSystems to use our own
  `ensure_sync` for all sync methods.
  '''
  def __enter__(self):
    if getattr(self, '__aenter__', None) is not None:
      return ensure_sync(self.__aenter__())
    elif getattr(super(), '__enter__', None) is not None:
      return super().__enter__()
    else:
      raise NotImplementedError

  def __exit__(self, *args):
    if getattr(self, '__aexit__', None) is not None:
      return ensure_sync(self.__aexit__(*args))
    elif getattr(super(), '__exit__', None) is not None:
      return super().__exit__(*args)
    else:
      raise NotImplementedError

  def mkdir(self, *args, **kwargs):
    if getattr(self, '_mkdir', None) is not None:
      return ensure_sync(self._mkdir(*args, **kwargs))
    elif getattr(super(), 'mkdir', None) is not None:
      return super().mkdir(*args, **kwargs)
    else:
      raise NotImplementedError
  
  def makedirs(self, *args, **kwargs):
    if getattr(self, '_makedirs', None) is not None:
      return ensure_sync(self._makedirs(*args, **kwargs))
    elif getattr(super(), 'makedirs', None) is not None:
      return super().makedirs(*args, **kwargs)
    else:
      raise NotImplementedError
  
  def rmdir(self, *args, **kwargs):
    if getattr(self, '_rmdir', None) is not None:
      return ensure_sync(self._rmdir(*args, **kwargs))
    elif getattr(super(), 'rmdir', None) is not None:
      return super().rmdir(*args, **kwargs)
    else:
      raise NotImplementedError
  
  def rm_file(self, *args, **kwargs):
    if getattr(self, '_rm_file', None) is not None:
      return ensure_sync(self._rm_file(*args, **kwargs))
    elif getattr(super(), 'rm_file', None) is not None:
      return super().rm_file(*args, **kwargs)
    else:
      raise NotImplementedError
  
  def rm(self, *args, **kwargs):
    if getattr(self, '_rm', None) is not None:
      return ensure_sync(self._rm(*args, **kwargs))
    elif getattr(super(), 'rm', None) is not None:
      return super().rm(*args, **kwargs)
    else:
      raise NotImplementedError
  
  def mv(self, *args, **kwargs):
    if getattr(self, '_mv', None) is not None:
      return ensure_sync(self._mv(*args, **kwargs))
    elif getattr(super(), 'mv', None) is not None:
      return super().mv(*args, **kwargs)
    else:
      raise NotImplementedError
  
  def cp_file(self, *args, **kwargs):
    if getattr(self, '_cp_file', None) is not None:
      return ensure_sync(self._cp_file(*args, **kwargs))
    elif getattr(super(), 'cp_file', None) is not None:
      return super().cp_file(*args, **kwargs)
    else:
      raise NotImplementedError
  
  def cat_file(self, *args, **kwargs):
    if getattr(self, '_cat_file', None) is not None:
      return ensure_sync(self._cat_file(*args, **kwargs))
    elif getattr(super(), 'cat_file', None) is not None:
      return super().cat_file(*args, **kwargs)
    else:
      raise NotImplementedError
  
  def put_file(self, *args, **kwargs):
    if getattr(self, '_put_file', None) is not None:
      return ensure_sync(self._put_file(*args, **kwargs))
    elif getattr(super(), 'put_file', None) is not None:
      return super().put_file(*args, **kwargs)
    else:
      raise NotImplementedError
  
  def get_file(self, *args, **kwargs):
    if getattr(self, '_get_file', None) is not None:
      return ensure_sync(self._get_file(*args, **kwargs))
    elif getattr(super(), 'get_file', None) is not None:
      return super().get_file(*args, **kwargs)
    else:
      raise NotImplementedError
  
  def exists(self, *args, **kwargs):
    if getattr(self, '_exists', None) is not None:
      return ensure_sync(self._exists(*args, **kwargs))
    elif getattr(super(), 'exists', None) is not None:
      return super().exists(*args, **kwargs)
    else:
      raise NotImplementedError
  
  def info(self, *args, **kwargs):
    if getattr(self, '_info', None) is not None:
      return ensure_sync(self._info(*args, **kwargs))
    elif getattr(super(), 'info', None) is not None:
      return super().info(*args, **kwargs)
    else:
      raise NotImplementedError
  
  def ls(self, *args, **kwargs):
    if getattr(self, '_ls', None) is not None:
      return ensure_sync(self._ls(*args, **kwargs))
    elif getattr(super(), 'ls', None) is not None:
      return super().ls(*args, **kwargs)
    else:
      raise NotImplementedError
  
  def glob(self, *args, **kwargs):
    if getattr(self, '_glob', None) is not None:
      return ensure_sync(self._glob(*args, **kwargs))
    elif getattr(super(), 'glob', None) is not None:
      return super().glob(*args, **kwargs)
    else:
      raise NotImplementedError



import os
import sys
import json
import traceback
import logging
logger = logging.getLogger()

from appyter.execspec.spec import AbstractExecutor
from appyter.ext.asyncio.subprocess import sh
from appyter.ext.json import async_json_loads

class SubprocessExecutor(AbstractExecutor):
  ''' Run executions in a subprocess
  '''
  protocol = 'subprocess'

  async def _submit(self, **job):
    async for msg, done in sh(
      sys.executable, '-u', '-m',
      'appyter', 'nbexecute',
      '-w', f"storage://{job['cwd']}",
      '-s', 'file:///dev/stdout',
      '--data-dir', job['storage'],
      '--fuse=true',
      job['ipynb'],
      stderr=sys.stderr,
      env=dict(
        os.environ,
        { k: v for k, v in os.environ.items() if not k.startswith('APPYTER_') },
        PYTHONPATH=':'.join(sys.path),
      ),
    ):
      yield msg, done

  async def _run(self, **job):
    yield dict(type='status', data=f"Launching subprocess...")
    async for msg, done in self._submit(**job):
      if not done:
        try: yield await async_json_loads(msg)
        except: logger.warning(traceback.format_exc())
    if msg != 0:
      yield dict(type='error', data=f"Subprocess exited with error code")



import os
import asyncio
import traceback
import logging
logger = logging.getLogger(__name__)

from appyter.render.flask_app.prepare import _prepare_data, _prepare_request, _prepare_results, _prepare_storage, _prepare_executor
from appyter.render.flask_app.socketio import socketio
from appyter.render.flask_app.room_manager import enter_room, find_room, room_lock
from appyter.ext.uuid import sanitize_sha1sum, generate_uuid
from appyter.ext.urllib import join_url

# construct/join a notebook
@socketio.on('submit')
async def submit(sid, data):
  async with socketio.session(sid) as sess:
    config = sess['config']
    request_url = sess['request_url']
    executor = sess['executor']
  #
  if type(data) == str:
    data = dict(_id=data)
  elif type(data) != dict:
    raise Exception('Unrecognized data type')
  #
  if '_id' in data:
    instance_id = sanitize_sha1sum(data['_id'])
    assert instance_id is not None, 'Invalid session id'
    data.update(_config=config)
  else:
    data = dict(_config=config)
    data.update(await _prepare_data(data))
    data.update(await _prepare_request(data))
    instance_id = await _prepare_results(data)
  #
  if 'catalog-integration' in data['_config']['EXTRAS']:
    # if you execute the notebook, it should get registered
    #  we'll omit metadata which would already be captured
    #  when the instance was created
    try:
      from appyter.extras.catalog_integration.notebooks import InstanceInfo, add_instance
      await add_instance(
        InstanceInfo(instance=instance_id),
        auth=data.get('_auth'),
        config=data.get('_config'),
      )
    except:
      logger.warning(traceback.format_exc())
  #
  if await find_room(instance_id):
    await socketio.emit('status', 'Joining existing execution...', to=sid)
    await enter_room(sid, instance_id)
  else:
    try:
      storage = await _prepare_storage(data)
      async with _prepare_executor(data, executor) as executor:
        async with room_lock(instance_id):
          await enter_room(sid, instance_id)
          try:
            await socketio.emit('status', 'Submitting execution...', to=instance_id)
            job = dict(
              cwd=f"output/{instance_id}",
              ipynb=os.path.basename(config['IPYNB']),
              session=instance_id,
              id=generate_uuid(),
              url=join_url(request_url, instance_id),
              storage=storage,
              debug=config['DEBUG'],
            )
            async for msg in executor._run(**job):
              await socketio.forward(None, dict(event=msg['type'], data=msg['data'], to=instance_id))
          except asyncio.CancelledError:
            raise
          except Exception:
            logger.error(traceback.format_exc())
            await socketio.emit('error', 'An unhandled executor error occurred, please try again later.', to=sid)
    except asyncio.CancelledError:
      raise
    except:
      logger.error(traceback.format_exc())
      await socketio.emit('error', 'An unhandled initialization error occurred, please try again later.', to=sid)
      raise



from fsspec import AbstractFileSystem

class ComposableAbstractFileSystem:
  ''' Expansion of AbstractFileSystem's json serialization methods to support composition
  by calling to_json() on any filesystems in storage_args during serialization,
  and loading them in during deserialization.
  '''
  def to_json(self):
    import json
    cls = type(self)
    cls = '.'.join((cls.__module__, cls.__name__))
    proto = (
      self.protocol[0]
      if isinstance(self.protocol, (tuple, list))
      else self.protocol
    )
    return json.dumps(
      dict(
        **{"cls": cls, "protocol": proto, "args": self.storage_args},
        **{
          k: json.loads(v.to_json()) if isinstance(v, AbstractFileSystem) else v
          for k, v in self.storage_options.items()
        },
      )
    )

  @staticmethod
  def from_json(blob):
    import json
    from fsspec.registry import _import_class, get_filesystem_class
    dic = json.loads(blob)
    protocol = dic.pop("protocol")
    try:
      cls = _import_class(dic.pop("cls"))
    except (ImportError, ValueError, RuntimeError, KeyError, AttributeError):
      cls = get_filesystem_class(protocol)
    if cls.from_json == ComposableAbstractFileSystem.from_json:
      return cls(*dic.pop("args", ()), **{
        k: ComposableAbstractFileSystem.from_json(json.dumps(v)) if type(v) == dict and 'cls' in v else v
        for k, v in dic.items()
      })
    else:
      return cls.from_json(blob)



import os
import sys
import click
import fsspec
import asyncio
import pathlib
import datetime
import traceback
import logging
logger = logging.getLogger(__name__)

from appyter import __version__
from appyter.cli import cli
from appyter.ext.emitter import json_emitter_factory
from appyter.ext.fsspec.core import url_to_chroot_fs
from appyter.ext.asyncio.helpers import ensure_async_contextmanager, ensure_sync
from appyter.ext.nbclient import NotebookClientIOPubHook
from appyter.parse.nb import nb_from_ipynb_io, nb_to_ipynb_io, nb_to_json
from appyter.ext.click import click_option_setenv, click_argument_setenv
from appyter.ext.urllib import URI

def cell_is_code(cell):
    return cell.get('cell_type') == 'code'

def cell_has_error(cell):
    for output in cell['outputs']:
        if output['output_type'] == 'error':
            return True
    return False

def iopub_hook_factory(nb, emit):
  async def iopub_hook(cell, cell_index):
    await emit({ 'type': 'cell', 'data': [cell, cell_index] })
  return iopub_hook

async def nbexecute_async(ipynb='', emit=json_emitter_factory(sys.stdout), cwd='', subscribe=None, fuse=False):
  logger.info('starting')
  assert callable(emit), 'Emit must be callable'
  with fsspec.open(str(URI(cwd).join(ipynb)), 'r') as fr:
    nb = nb_from_ipynb_io(fr)
  #
  if 'appyter' not in nb.metadata:
    logger.warning('detected legacy format, upgrading..')
    nb.metadata['appyter'] = {
      'nbconstruct': {
        'version': 'unknown',
      }
    }
  #
  if 'nbexecute' not in nb.metadata['appyter']:
    nb.metadata['appyter']['nbexecute'] = {
      'version': __version__,
    }
  #
  if 'execution_info' in nb.metadata:
    logger.warning('detected legacy format, upgrading..')
    nb.metadata['appyter']['nbexecute'].update(
      version='unknown',
      started=nb.metadata['execution_info'].get('started'),
      completed=nb.metadata['execution_info'].get('completed'),
    )
    del nb.metadata['execution_info']
  #
  if 'completed' in nb.metadata['appyter']['nbexecute']:
    await emit({ 'type': 'error', 'data': f"Execution already completed at {nb.metadata['appyter']['nbexecute']['completed']}" })
    return
  elif 'started' in nb.metadata['appyter']['nbexecute']:
    await emit({ 'type': 'error', 'data': f"Execution already started at {nb.metadata['appyter']['nbexecute']['started']}" })
    return
  #
  await emit({ 'type': 'status', 'data': 'Starting' })
  state = dict(progress=0, status='Starting')
  #
  try:
    files = nb.metadata['appyter']['nbconstruct'].get('files')
    logger.debug(f"{cwd=} {files=}")
    async with ensure_async_contextmanager(url_to_chroot_fs(cwd, pathmap=files)) as fs:
      async with ensure_async_contextmanager(fs.mount(fuse=fuse)) as mnt:
        # setup execution_info with start time
        nb.metadata['appyter']['nbexecute']['started'] = datetime.datetime.now().replace(tzinfo=datetime.timezone.utc).isoformat()
        with (mnt/ipynb).open('w') as fw:
          nb_to_ipynb_io(nb, fw)
        #
        logger.info('initializing')
        if callable(subscribe):
          await subscribe(lambda: dict(nb=nb_to_json(nb), **state))
        #
        try:
          iopub_hook = iopub_hook_factory(nb, emit)
          client = NotebookClientIOPubHook(
            nb,
            allow_errors=True,
            timeout=None,
            kernel_name='python3',
            resources={ 'metadata': {'path': str(mnt) } },
            iopub_hook=iopub_hook,
          )
          await emit({ 'type': 'nb', 'data': nb_to_json(nb) })
          try:
            async with client.async_setup_kernel(
              env=dict(
                { k: v for k, v in os.environ.items() if not k.startswith('APPYTER_') },
                PYTHONPATH=':'.join(sys.path),
              ),
            ):
              logger.info('executing')
              state.update(status='Executing...', progress=0)
              await emit({ 'type': 'status', 'data': state['status'] })
              await emit({ 'type': 'progress', 'data': state['progress'] })
              client.set_widgets_metadata()
              n_cells = len(nb.cells)
              exec_count = 1
              for index, cell in enumerate(nb.cells):
                logger.debug(f"executing cell {index}")
                cell = await client.async_execute_cell(
                  cell, index,
                  execution_count=exec_count,
                )
                if cell_is_code(cell):
                  if cell_has_error(cell):
                    raise Exception('Cell execution error on cell %d' % (exec_count))
                  exec_count += 1
                if index < n_cells-1:
                  state['progress'] = index + 1
                  await emit({ 'type': 'progress', 'data': state['progress'] })
                else:
                  state['status'] = 'Success'
                  await emit({ 'type': 'status', 'data': state['status'] })
          finally:
            client.set_widgets_metadata()
        except asyncio.CancelledError:
          logger.info('cancelled')
          raise
        except Exception as e:
          logger.info(f"execution error: {traceback.format_exc()}")
          await emit({ 'type': 'error', 'data': str(e) })
        # Save execution completion time
        logger.info('saving')
        nb.metadata['appyter']['nbexecute']['completed'] = datetime.datetime.now().replace(tzinfo=datetime.timezone.utc).isoformat()
        # save additional files
        # TODO: in the future we should individual files and include the original urls here
        nb.metadata['appyter']['nbexecute']['files'] = {
          path: path
          for path in (
            str(p.relative_to(mnt))
            for p in mnt.rglob('*')
            if p.is_file()
          )
          if path not in (files.keys()|{ipynb})
        }
        #
        with (mnt/ipynb).open('w') as fw:
          nb_to_ipynb_io(nb, fw)
        #
        logger.info('finalized')
  except asyncio.CancelledError:
    logger.info(f"outer cancelled")
    raise
  except Exception as e:
    logger.error(traceback.format_exc())
    if state['status'] != 'Success':
      await emit({ 'type': 'error', 'data': f"Error occured while executing" })
    else:
      await emit({ 'type': 'error', 'data': f"Error occured while finalizing" })
    raise
  finally:
    logger.info('complete')
  #

@cli.command(help='Execute a jupyter notebook on the command line asynchronously')
@click.option('-s', type=str, metavar='URI', default='file:///dev/stdout', help='Status stream')
@click.option('-w', type=str, metavar='DIR', default='storage://', help='Working directory')
@click_option_setenv('--data-dir', envvar='APPYTER_DATA_DIR', default=None, help='The directory that storage:// uris correspond to')
@click_option_setenv('--fuse', envvar='APPYTER_FUSE', default=False, help='Use fuse for execution')
@click_argument_setenv('ipynb', envvar='APPYTER_IPYNB')
def nbexecute(ipynb, s=None, w=None, fuse=False, data_dir=None):
  from appyter.ext.emitter import url_to_emitter
  from appyter.ext.asyncio.event_loop import with_event_loop
  from appyter.ext.fsspec.storage import ensure_storage
  if data_dir is None:
    p = pathlib.Path(ipynb)
    data_dir = str(p.parent)
    ipynb = p.name
  with with_event_loop():
    with ensure_sync(url_to_emitter(s)) as emitter:
      with ensure_sync(ensure_storage(data_dir)):
        ensure_sync(nbexecute_async(
          ipynb=ipynb,
          emit=emitter,
          cwd=w,
          fuse=fuse,
        ))



''' In the future, these will be independent of the profiles, injecting whatever needs to be.
For now, they're branches in the profile code.
'''

from appyter.cli import cli

@cli.command(help='List the available extra feature flags')
def list_extras():
  print('catalog-integration Add various appyter-catalog-integrated specific features')
  print('hide-code           Hide code by default')
  print('storage-file-field  Enable StorageFileField, WARNING: security risk, not for use in public environment')
  print('toc                 Add a sticky table of contents to the rendered notebook')
  print('toggle-code         Add a button for toggling code visibility')


from appyter.ext.asyncio.helpers import ensure_sync
import logging

logger = logging.getLogger(__name__)

async def run_app(config):
  import os
  import sys
  import asyncio
  return await asyncio.create_subprocess_exec(
    sys.executable, '-u', '-m', 'appyter',
    f"--socket={config['HOST']}:{config['PORT']}",
    stdout=sys.stdout,
    stderr=sys.stderr,
    env=os.environ,
  )

async def app_runner(emitter, config):
  import asyncio
  from appyter.ext.subprocess import interrupt
  state_lock = asyncio.Lock()
  state = dict(proc=await run_app(config))
  #
  @emitter.on('reload')
  async def reload(changes):
    logger.info(f"Reload {changes}")
    async with state_lock:
      if 'proc' in state:
        proc = state.pop('proc')
        interrupt(proc)
        state['proc'] = await run_app(config)
  #
  @emitter.on('quit')
  async def quit():
    logger.info('Stopping app..')
    async with state_lock:
      proc = state.pop('proc')
      interrupt(proc)

async def app_messager(emitter, config):
  import asyncio
  from appyter.ext.urllib import join_slash
  from appyter.ext.socketio import AsyncClient
  from appyter.ext.asyncio.try_n_times import async_try_n_times
  async with AsyncClient() as sio:
    #
    @emitter.on('livereload')
    async def livereload(changes):
      logger.info(f"LiveReload {changes}")
      await sio.emit('livereload', {})
    #
    @emitter.on('quit')
    async def quit():
      logger.info('Disconnecting..')
      await sio.disconnect()
    #
    origin = f"http://{config['HOST']}:{config['PORT']}"
    path = join_slash(config['PREFIX'], "socket.io")
    logger.info(f"Connecting to appyter server at {origin}{path}...")
    await asyncio.sleep(1)
    await async_try_n_times(3, sio.connect, origin, socketio_path=path)
    await sio.wait()

async def file_watcher(emitter, evt, path, **kwargs):
  from watchgod import awatch
  logger.info(f"Watching {path} for {evt}...")
  async for changes in awatch(path, **kwargs):
    await emitter.emit(evt, changes=changes)

@ensure_sync
async def serve(app_path, **kwargs):
  import os
  import asyncio
  import appyter
  from appyter.ext.asyncio.event_emitter import EventEmitter
  from appyter.ext.watchgod.watcher import GlobWatcher
  from appyter.context import get_env
  config = get_env(**kwargs)
  emitter = EventEmitter()
  tasks = []
  # run the app and reload it when necessary
  tasks.append(asyncio.create_task(app_runner(emitter, config)))
  # the underlying appyter library
  if config['WATCH']:
    if os.environ.get('APPYTER_DEVELOPMENT'):
      tasks.append(asyncio.create_task(
        file_watcher(emitter, 'reload', appyter.__path__[0],
          watcher_cls=GlobWatcher,
          watcher_kwargs=dict(
            include_dir_glob=['*'],
            include_file_glob=['*.py'],
            exclude_dir_glob=[],
            exclude_file_glob=[],
          ),
        )
      ))
      # the underlying appyter library's templates/ipynb/staticfiles/...
      tasks.append(asyncio.create_task(
        file_watcher(emitter, 'livereload', appyter.__path__[0],
          watcher_cls=GlobWatcher,
          watcher_kwargs=dict(
            include_dir_glob=['*'],
            include_file_glob=['*'],
            exclude_dir_glob=[],
            exclude_file_glob=['*.py'],
          ),
        )
      ))
    # the appyter itself's filters/blueprints
    tasks.append(asyncio.create_task(
      file_watcher(emitter, 'reload', config['CWD'],
        watcher_cls=GlobWatcher,
        watcher_kwargs=dict(
          include_dir_glob=['filters', 'blueprints'],
          include_file_glob=['*.py'],
          exclude_dir_glob=[],
          exclude_file_glob=[],
        ),
      )
    ))
    # the appyter itself's templates/ipynb/staticfiles/...
    tasks.append(asyncio.create_task(
      file_watcher(emitter, 'livereload', config['CWD'],
        watcher_cls=GlobWatcher,
        watcher_kwargs=dict(
          include_dir_glob=['*'],
          include_file_glob=['*'],
          exclude_dir_glob=[config['DATA_DIR']],
          exclude_file_glob=['*.py'],
        ),
      )
    ))
  tasks.append(asyncio.create_task(app_messager(emitter, config)))
  exit_code = 0
  try:
    await asyncio.Event().wait()
  except asyncio.CancelledError:
    pass
  except:
    exit_code = 1
  finally:
    await emitter.emit('quit')
    await emitter.flush()
    await emitter.clear()
    for task in tasks:
      try:
        task.cancel()
        await task
      except asyncio.CancelledError:
        pass
  return exit_code



from appyter.ext.urllib import URI
from appyter.ext.asyncio.helpers import ensure_sync

class AbstractExecutor:
  protocol = None

  def __init__(self, url=None, **kwargs) -> None:
    self.url = url
    self.executor_options = kwargs
  
  @classmethod
  def parse(cls, url) -> dict:
    url = URI(url)
    return dict(url=str(url.with_fragment(None)), **url.fragment_query_ex)

  async def __aenter__(self):
    return self

  def __enter__(self):
    return ensure_sync(self.__aenter__())

  async def __aexit__(self, type, value, traceback):
    pass

  def __exit__(self, type, value, traceback):
    return ensure_sync(self.__aexit__(type, value, traceback))

  async def _run(self, **job):
    raise NotImplementedError

  def run(self, **job):
    return ensure_sync(self._run(**job))



import os
import sys
import json
import logging
import asyncio
import traceback

logger = logging.getLogger(__name__)

from appyter.execspec.spec import AbstractExecutor
from appyter.ext.asyncio.subprocess import sh
from appyter.ext.dict import dict_merge
from appyter.ext.json import async_json_loads

class DockerExecutor(AbstractExecutor):
  ''' Run executions in a docker job
  Example uri:
  docker::maayanlab/myimage:latest?flags.v=/mydatadir:/data&args.data-dir=/data
  '''
  protocol = 'docker'

  async def __aenter__(self):
    self._args = [
      'docker', 'run',
      *(
        f"--{k}={v}" if len(k) > 1 else f"-{k}{v}"
        for k, v in dict_merge({
          'device': '/dev/fuse',
          'cap-add': 'SYS_ADMIN',
          'security-opt': 'apparmor:unconfined',
        }, **self.executor_options.get('flags', {})).items()
      ),
    ]
    try:
      proc = await asyncio.create_subprocess_exec(*[
        'docker', 'inspect', os.environ['HOSTNAME'],
      ], stdout=asyncio.subprocess.PIPE)
      proc_stdout, _ = await proc.communicate()
      conf = await async_json_loads(proc_stdout.decode())
      self._args += [
        f"--network={conf[0]['HostConfig']['NetworkMode']}"
      ]
    except:
      logger.warn('Not a docker container')
    self._args.append(self.url)
    return await super().__aenter__()

  async def _submit(self, **job):
    full_args = [
      *self._args,
      'appyter', 'nbexecute',
      *(
        f"--{k}={v}" if len(k) > 1 else f"-{k}{v}"
        for k, v in dict_merge(
          {
            'w': f"storage://{job['cwd']}",
            's': 'file:///dev/stdout',
            'fuse': 'true',
            'data-dir': job['storage'],
          },
          **self.executor_options.get('args', {})
        ).items()
      ),
      job['ipynb'],
    ]
    async for msg, done in sh(*full_args, stderr=sys.stderr):
      yield msg, done

  async def _run(self, **job):
    yield dict(type='status', data=f"Launching container...")
    async for msg, done in self._submit(**job):
      if not done:
        try: yield await async_json_loads(msg)
        except asyncio.CancelledError: raise
        except: logger.warning(traceback.format_exc())
    if msg != 0:
      yield dict(type='error', data=f"Container exited with error code")
